/*
 * Copyright (c) 2023 - 2024. KeyWare.Co.Ltd All rights reserved.
 * 项目名称：28所 C++ 信息安全性设计准则
 * 项目描述：用于检查C++源代码的安全性设计准则的Sonarqube插件
 * 版权说明：本软件属北京关键科技股份有限公司所有，在未获得北京关键科技股份有限公司正式授权情况下，任何企业和个人，不能获取、阅读、安装、传播本软件涉及的任何受知识产权保护的内容。
 */
package com.keyware.sonar.cxx.rules.checkers;

import com.sonar.cxx.sslr.api.AstNode;
import com.sonar.cxx.sslr.api.Grammar;
import org.sonar.api.utils.log.Logger;
import org.sonar.api.utils.log.Loggers;
import org.sonar.check.Priority;
import org.sonar.check.Rule;
import org.sonar.cxx.parser.CxxGrammarImpl;
import org.sonar.cxx.squidbridge.api.CheckMessage;
import org.sonar.cxx.squidbridge.api.SourceCode;
import org.sonar.cxx.visitors.MultiLocatitionSquidCheck;

import javax.annotation.Nonnull;
import java.util.Set;


/**
 * 应通过用户名口令、数字证书等其他手段对主机身份进行鉴别
 * <p>仅检测代码中有没有用户名口令和数字证书</p>
 *
 * @author GuoXin
 * @date 2024/1/24
 */
@Rule(key = "HostIdentityVerifyChecker", name = "程序中未识别到通过用户名口令或者数字证书验证等安全认证逻辑", description = "应通过用户名口令、数字证书等其他手段对主机身份进行鉴别", priority = Priority.MAJOR, tags = {"28suo"})
public class HostIdentityVerifyChecker extends MultiLocatitionSquidCheck<Grammar> {
    private static final Logger LOG = Loggers.get(HostIdentityVerifyChecker.class);
    private SourceCode project;

    private boolean hasUsername = false;
    private boolean hasPassword = false;
    private boolean hasCertificate = false;

    private final Set<String> SSL_FUNCTIONS = Set.of(
            "SSL_use_certificate_file",
            "SSL_use_PrivateKey_file",
            "SSL_use_RSAPrivateKey_file",
            "SSL_CTX_use_certificate_file",
            "SSL_CTX_use_RSAPrivateKey_file",
            "SSL_CTX_use_PrivateKey_file",
            "SSL_CTX_use_certificate_chain_file",
            "SSL_use_certificate_chain_file",
            "SSL_library_init",
            "PEM_read_X509",
            "d2i_X509_bio",
            "X509_verify",
            "X509_check_issued",
            "X509_get_notBefore",
            "X509_get_notAfter",
            "X509_get_subject_name",
            "X509_get_issuer_name",
            "X509_get_pubkey",
            "X509_STORE_CTX_init",
            "X509_STORE_add_cert",
            "X509_verify_cert"
    );

    @Override
    public void init() {
        LOG.info("[KeyWare][C++检查] >> 应通过用户名口令、数字证书等其他手段对主机身份进行鉴别");
        if (project == null) {
            project = getContext().peekSourceCode();
        }
        subscribeTo(CxxGrammarImpl.functionDefinition, CxxGrammarImpl.simpleDeclaration, CxxGrammarImpl.postfixExpression);
    }


    @Override
    public void visitNode(@Nonnull AstNode astNode) {
        if ((hasUsername && hasPassword) || hasCertificate) {
            return;
        }

        if (astNode.is(CxxGrammarImpl.functionDefinition)) {
            visitFunctionDefinition(astNode);
        } else if (astNode.is(CxxGrammarImpl.simpleDeclaration)) {
            visitSimpleDeclaration(astNode);
        } else if (astNode.is(CxxGrammarImpl.postfixExpression)) {
            visitPostfixExpression(astNode);
        }
    }

    private void visitFunctionDefinition(@Nonnull AstNode astNode) {
        astNode.getDescendants(CxxGrammarImpl.parameterDeclaration).forEach(this::checkDeclarationNode);
    }

    private void visitSimpleDeclaration(@Nonnull AstNode astNode) {
        astNode.getDescendants(CxxGrammarImpl.declarator).forEach(this::checkDeclarationNode);
    }

    private void visitPostfixExpression(@Nonnull AstNode astNode) {
        var nameNode = astNode.getFirstDescendant(CxxGrammarImpl.declaratorId);
        if (nameNode == null) {
            nameNode = astNode.getFirstDescendant(CxxGrammarImpl.className);
        }
        if (nameNode != null) {
            var nameStr = nameNode.getTokenValue();
            if (SSL_FUNCTIONS.contains(nameStr)) {
                hasCertificate = true;
            }
        }
    }

    private void checkDeclarationNode(@Nonnull AstNode astNode) {
        var paramNode = astNode.getFirstDescendant(CxxGrammarImpl.declaratorId);
        if (paramNode != null) {
            var paramName = paramNode.getTokenValue();
            paramName = paramName.toLowerCase().replace("-", "").replace("_", "");
            if ("username".equals(paramName)) {
                hasUsername = true;
            } else if ("password".equals(paramName)) {
                hasPassword = true;
            }
        }
    }

    /**
     * 当访问器销毁时执行，代表所有文件已经被访问完毕，可以进行最后的处理。
     */
    @Override
    public void destroy() {
        if (!(hasPassword && hasUsername) && !hasCertificate) {
            project.log(new CheckMessage(this, "应通过用户名口令、数字证书等其他手段对主机身份进行鉴别"));
        }
    }
}
