/*
 * Copyright (c) 2023 - 2024. KeyWare.Co.Ltd All rights reserved.
 * 项目名称：信息安全性设计准则检查插件
 * 项目描述：用于检查源代码的安全性设计准则的Sonarqube插件
 * 版权说明：本软件属北京关键科技股份有限公司所有，在未获得北京关键科技股份有限公司正式授权情况下，任何企业和个人，不能获取、阅读、安装、传播本软件涉及的任何受知识产权保护的内容。
 */
package com.keyware.sonar.java.rules.checkers;


/**
 * 要求用户使用具有足够复杂度的口令。
 *
 * @author WuHaoyang
 * @date 2024/1/12
 */
import org.sonar.check.Rule;
import org.sonar.plugins.java.api.IssuableSubscriptionVisitor;
import org.sonar.plugins.java.api.tree.*;

import java.util.*;

@Rule(key = "PasswordRegexCheck")
public class PasswordRegexCheck extends IssuableSubscriptionVisitor {

    private static final String MATCHER_METHOD = "matcher";
    private static final String PASSWORD_PARAMETER = "password";

    @Override
    public List<Tree.Kind> nodesToVisit() {
        return Collections.singletonList(Tree.Kind.METHOD);
    }

    @Override
    public void visitNode(Tree tree) {
        if (tree.is(Tree.Kind.METHOD)) {
            MethodTree methodTree = (MethodTree) tree;
            checkPasswordValidation(methodTree);
        }
    }

    private void checkPasswordValidation(MethodTree methodTree) {
        boolean hasPasswordValidation = false;

        for (StatementTree statement : methodTree.block().body()) {
            if (statement.is(Tree.Kind.VARIABLE)) {
                VariableTree variableTree = (VariableTree) statement;
                if (variableTree.initializer() != null && variableTree.initializer().is(Tree.Kind.METHOD_INVOCATION)) {
                    MethodInvocationTree methodInvocationTree = (MethodInvocationTree) variableTree.initializer();
                    String methodName = methodInvocationTree.methodSymbol().name();
                    if (MATCHER_METHOD.equals(methodName)) {
                        hasPasswordValidation = hasPasswordValidation || hasPasswordParameter(methodInvocationTree.arguments());
                    }
                }
            }
        }

        if (!hasPasswordValidation) {
            //如果没有发现密码验证，报告问题
            System.out.println("未对口令进行复杂度验证"+methodTree.simpleName());
            reportIssue(methodTree.simpleName(), "未对口令进行复杂度验证");
        }
    }

    private boolean hasPasswordParameter(List<ExpressionTree> arguments) {
        for (ExpressionTree argument : arguments) {
            if (argument.is(Tree.Kind.IDENTIFIER)) {
                IdentifierTree identifier = (IdentifierTree) argument;
                if (PASSWORD_PARAMETER.equalsIgnoreCase(identifier.name())) {
                    //检查标识符是否有密码验证
                    return hasPasswordValidationInMethod(identifier.name());
                }
            }
        }

        return false;
    }

    private boolean hasPasswordValidationInMethod(String paramName) {
        //参数名是'password'，已经存在
        return PASSWORD_PARAMETER.equalsIgnoreCase(paramName);
    }

}
